/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2014 Ryan Zezeski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */


/*
 * Inspired by both Stevens's rtsignals/test1 and Anton Salikhmetov's
 * 2007 LKML post.
 *
 * http://www.kohala.com/start/unpv22e/unpv22e.html
 * https://lkml.org/lkml/2007/7/10/517
 *
 */
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/*
 * The seq struct and sh handler are used to record the order in which
 * the signals are delivered to the parent process.  When a signal is
 * delivered the handler records it in seq and then increments the
 * index.
 */
struct seq {
	int sig, num;
};

static struct seq *seqp;

void
sh(int sig, siginfo_t *info, void *unused)
{
	static int i;
	seqp[i].num = info->si_value.sival_int;
	seqp[i++].sig = sig;
}

/*
 * Verify the delivery order of realtime (RT) signals.  RT signals were
 * introduced in POSIX.1b and differ from regular signals in several key
 * ways:
 *
 * 1. Multiple instances of the same signal are queued and must be
 * delivered in FIFO order.
 *
 * 2. If multiple different RT signals are pending then the lower numbered
 * RT signals are delivered first.  I.e., RT signals have a notion of
 * priority.
 *
 * See sections 2.4.2 (Realtime Signal Generation and Delivery) and B.2.4
 * of the POSIX.1-2008 standard for more information.
 */
static int
rt_delivery(int no_defer)
{
	int i, j, sig, sival, sigend, numsig, fail = 0;
	pid_t p = getpid(), chld;
	struct sigaction sa;
	sigset_t ss, ssold;
	union sigval sv;

	/*
	 * Range of signals to send.  Each signal is sent 2 times to verify FIFO
	 * order.
	 */
	sigend = SIGRTMIN + (sysconf(_SC_SIGQUEUE_MAX) / 2);
	sigend = sigend > SIGRTMAX ? SIGRTMAX : sigend;
	numsig = (sigend - SIGRTMIN) * 2;

	seqp = malloc(sizeof (struct seq) * numsig);

	/*
	 * Step 1:
	 *
	 * Install the signal handler (sh) on this process (the parent).
	 */
	sigemptyset(&ss);
	for (i = SIGRTMIN; i < sigend; i++) {
		sa.sa_sigaction = sh;
		sa.sa_flags = SA_SIGINFO;
		sa.sa_flags |= (no_defer ? SA_NODEFER : 0);
		sigemptyset(&sa.sa_mask);
		sigaction(i, &sa, NULL);
		sigaddset(&ss, i);
	}

	/*
	 * Step 2:
	 *
	 * Block all signals that are about to be sent to this process.
	 * This is needed to make sure all signals are pending at once.
	 * This is related to step 4.
	 */
	sigprocmask(SIG_BLOCK, &ss, &ssold);

	/*
	 * Step 3:
	 *
	 * Fork a child process to send each RT signal 2 times to the
	 * parent process.  Remember, all signals being sent are RT
	 * signals and thus are subject to explicit ordering guarantees
	 * spelled out in POSIX.1b.  Lower numbered RT signals are
	 * _always_ delivered before higher RT signals.
	 */
	if ((chld = fork()) == 0) {
		for (i = SIGRTMIN; i < sigend; i++) {
			for (j = 0; j < 2; j++) {
				sv.sival_int = j;
				if (sigqueue(p, i, sv) == -1) {
					perror("problem generating signal");
					exit(1);
				}
			}
		}
		exit(0);
	}

	/*
	 * Step 4:
	 *
	 * Wait for forked process to finish sending all signals.  This is
	 * to make sure that all signals are pending before allowing the
	 * kernel to deliver them to the handler.  If you don't both wait
	 * and block then the handler will process the signals as they
	 * come in and thus we will not be testing the ordering
	 * guarentees.  Related to step 2.
	 */
	if (wait(NULL) != chld) {
		perror("problem waiting for child");
		exit(1);
	}
	sigprocmask(SIG_SETMASK, &ssold, NULL);

	/*
	 * Step 5:
	 *
	 * Verify that multiple pending RT signals are delivered in
	 * correct priority (lower signals first), and that multiple
	 * instances of the same RT signal are delivered in FIFO order.
	 *
	 * E.g. if SIGRTMIN is 42 then the handler should record this
	 * ordering:
	 *
	 * {42, 0}, {42, 1}, {42, 2}, {43, 0}, {43, 1}, {43, 2}, ...
	 */
	sig = SIGRTMIN;
	sival = 0;
	for (i = 0; i < numsig; i++) {
		if ((seqp[i].sig != sig) || seqp[i].num != sival) {
			fail = 1;
		}
		sig = ((i + 1) % 2) == 0 ? sig + 1 : sig;
		sival = ((i + 1) % 2) == 0 ? 0 : sival + 1;
	}

	if (fail) {
		sig = SIGRTMIN;
		sival = 0;
		fprintf(stderr, "ERROR: incorrect order\n");
		fprintf(stderr, "Excpted Actual\n");
		for (i = 0; i < numsig; i++) {
			fprintf(stderr, "{%2d,%d}  {%2d,%d}\n",
			    sig, sival, seqp[i].sig, seqp[i].num);

			sig = ((i + 1) % 2) == 0 ? sig + 1 : sig;
			sival = ((i + 1) % 2) == 0 ? 0 : sival + 1;
		}
		exit(1);
	}

	free(seqp);
	return (0);
}

static void
rt_delivery_defer_test(void)
{
	printf("verify delivery order of pending RT signals...\n");
	rt_delivery(0);
	printf("PASSED\n");
}

static void
rt_delivery_nodefer_test(void)
{
	printf("verify delivery order of pending RT signals w/ SA_NODEFER...\n");
	rt_delivery(1);
	printf("PASSED\n");
}

int
main(int argc, char *argv[])
{
	rt_delivery_defer_test();
	rt_delivery_nodefer_test();
	return (EXIT_SUCCESS);
}
